@@ -1,5 +1,5 @@
 import requests
-from typing import List, Optional
+from typing import List, Optional, Callable, Any
 
 
 from .models import User, Template, Gift, Campaign, CardResponse, CardsResponse, CampaignResponse, Card, Contact, Mailing
@@ -19,10 +19,66 @@ FIXED_COUNTRY = {
 
 class AMcardsClient:
     """Client for AMcards API."""
-    def __init__(self, access_token: str) -> None:
+    def __init__(self, access_token: str, oauth_config: Optional[dict] = None, callback: Optional[Callable[[str, str, int], Any]] = None) -> None:
+        """Client for AMcards API.
+
+        :param str access_token: Your AMcards access token. Generate one `here <https://amcards.com/user/generate-access-token/>`_.
+        :param Optional[dict] oauth_config: Your OAuth configuration, this is optional, but if you choose to include it, it must include all of the following keys:
+
+            .. code-block::
+
+                {
+                    'refresh_token': 'yourrefreshtoken',
+                    'expiration': 1671692131130,                    # Note that expiration is specified as a unix timestamp in ms.
+                    'client_id': 'yourapplicationclientid',
+                    'client_secret': 'yourapplicationclientsecret'
+                }
+
+        :param Optional[Callable[[str, str, int], Any]] callback: This function will be called by the client when you have ``oauth_config`` defined, and the client refreshes the ``access_token`` before making a request. This can be useful when you need perform some logic when the client refreshes the specified ``access_token`` (like updating your database to reflect the new access_token and refresh_token). ``callback`` is optional, but if you choose to include it, it needs to accept the following keyword arguments: ``access_token`` ``refresh_token`` ``expiration``. ``callback`` will be called by the client as follows: ``callback(access_token='newaccesstoken', refresh_token='newrefreshtoken', expiration=1671692135130)``, note that ``expiration`` is specified as a unix timestamp in ms.
+
+        """
         self._access_token = access_token
-        self.HEADERS = {
-            'Authorization': f'Bearer {access_token}',
+        self._oauth_config = oauth_config
+        self._callback = callback
+
+    def _token_expired(self) -> bool:
+        return self._oauth_config is not None and helpers.current_timestamp() >= self._oauth_config['expiration']
+
+    def _refresh_token(self) -> None:
+        # Refresh the tokens
+        payload = {
+            'grant_type': 'refresh_token',
+            'refresh_token': self._oauth_config['refresh_token'],
+            'client_id': self._oauth_config['client_id'],
+            'client_secret': self._oauth_config['client_secret'],
+        }
+        res = requests.post(url=f'{DOMAIN}/oauth2/token/', data=payload)
+
+        # Make sure the refresh was successful
+        if not res.ok:
+            raise exceptions.OAuthTokenRefreshError('Something went wrong when attempting to refresh AMcards access_token')
+
+        # Update the access_token, refresh_token, and expiration
+        res_json = res.json()
+        self._access_token = res_json['access_token']
+        self._oauth_config['refresh_token'] = res_json['refresh_token']
+        self._oauth_config['expiration'] = helpers.current_timestamp() + res_json['expires_in'] * 1000
+
+        # If specified, call the callback
+        if self._callback is not None:
+            self._callback(
+                access_token=self._access_token,
+                refresh_token=self._oauth_config['refresh_token'],
+                expiration=self._oauth_config['expiration'],
+            )
+
+    @property
+    def _HEADERS(self) -> dict:
+        if self._token_expired():
+            self._refresh_token()
+
+        return {
+            'Authorization': f'Bearer {self._access_token}',
         }
 
     def user(self) -> User:
@@ -34,7 +90,7 @@ class AMcardsClient:
         :raises AuthenticationError: When the client's ``access_token`` is invalid.
 
         """
-        res = requests.get(url=f'{DOMAIN}/.api/v1/user/', headers=self.HEADERS)
+        res = requests.get(url=f'{DOMAIN}/.api/v1/user/', headers=self._HEADERS)
         if not res.ok:
             raise exceptions.AuthenticationError('Access token provided to client is unauthorized')
 
@@ -53,7 +109,7 @@ class AMcardsClient:
         :raises AuthenticationError: When the client's ``access_token`` is invalid.
 
         """
-        res = requests.get(url=f'{DOMAIN}/.api/v1/template/', headers=self.HEADERS, params={'limit': limit, 'offset': skip})
+        res = requests.get(url=f'{DOMAIN}/.api/v1/template/', headers=self._HEADERS, params={'limit': limit, 'offset': skip})
         if not res.ok:
             raise exceptions.AuthenticationError('Access token provided to client is unauthorized')
 
@@ -72,7 +128,7 @@ class AMcardsClient:
         :raises AuthenticationError: When the client's ``access_token`` is invalid.
 
         """
-        res = requests.get(url=f'{DOMAIN}/.api/v1/template/{id}/', headers=self.HEADERS)
+        res = requests.get(url=f'{DOMAIN}/.api/v1/template/{id}/', headers=self._HEADERS)
         if not res.ok:
             if res.status_code in (403, 404):
                 raise exceptions.ForbiddenTemplateError('The template for the specified id either does not exist or is not owned by the client\'s user')
@@ -93,7 +149,7 @@ class AMcardsClient:
         :raises AuthenticationError: When the client's ``access_token`` is invalid.
 
         """
-        res = requests.get(url=f'{DOMAIN}/.api/v1/quicksendtemplate/', headers=self.HEADERS, params={'limit': limit, 'offset': skip})
+        res = requests.get(url=f'{DOMAIN}/.api/v1/quicksendtemplate/', headers=self._HEADERS, params={'limit': limit, 'offset': skip})
         if not res.ok:
             raise exceptions.AuthenticationError('Access token provided to client is unauthorized')
 
@@ -112,7 +168,7 @@ class AMcardsClient:
         :raises AuthenticationError: When the client's ``access_token`` is invalid.
 
         """
-        res = requests.get(url=f'{DOMAIN}/.api/v1/quicksendtemplate/{id}/', headers=self.HEADERS)
+        res = requests.get(url=f'{DOMAIN}/.api/v1/quicksendtemplate/{id}/', headers=self._HEADERS)
         if not res.ok:
             if res.status_code in (403, 404):
                 raise exceptions.ForbiddenTemplateError('The quicksend template for the specified id either does not exist or is not owned by the client\'s user')
@@ -133,7 +189,7 @@ class AMcardsClient:
         :raises AuthenticationError: When the client's ``access_token`` is invalid.
 
         """
-        res = requests.get(url=f'{DOMAIN}/.api/v1/campaign/', headers=self.HEADERS, params={'limit': limit, 'offset': skip})
+        res = requests.get(url=f'{DOMAIN}/.api/v1/campaign/', headers=self._HEADERS, params={'limit': limit, 'offset': skip})
         if not res.ok:
             raise exceptions.AuthenticationError('Access token provided to client is unauthorized')
 
@@ -152,7 +208,7 @@ class AMcardsClient:
         :raises AuthenticationError: When the client's ``access_token`` is invalid.
 
         """
-        res = requests.get(url=f'{DOMAIN}/.api/v1/campaign/{id}/', headers=self.HEADERS)
+        res = requests.get(url=f'{DOMAIN}/.api/v1/campaign/{id}/', headers=self._HEADERS)
         if not res.ok:
             if res.status_code in (403, 404):
                 raise exceptions.ForbiddenCampaignError('The drip campaign for the specified id either does not exist or is not owned by the client\'s user')
@@ -181,7 +237,7 @@ class AMcardsClient:
             'offset': skip,
         } | (filters or {})
 
-        res = requests.get(url=f'{DOMAIN}/.api/v1/card/', headers=self.HEADERS, params=params)
+        res = requests.get(url=f'{DOMAIN}/.api/v1/card/', headers=self._HEADERS, params=params)
         if not res.ok:
             raise exceptions.AuthenticationError('Access token provided to client is unauthorized')
 
@@ -200,7 +256,7 @@ class AMcardsClient:
         :raises AuthenticationError: When the client's ``access_token`` is invalid.
 
         """
-        res = requests.get(url=f'{DOMAIN}/.api/v1/card/{id}/', headers=self.HEADERS)
+        res = requests.get(url=f'{DOMAIN}/.api/v1/card/{id}/', headers=self._HEADERS)
         if not res.ok:
             if res.status_code in (403, 404):
                 raise exceptions.ForbiddenCardError('The card for the specified id either does not exist or is not owned by the client\'s user')
@@ -221,7 +277,7 @@ class AMcardsClient:
         :raises AuthenticationError: When the client's ``access_token`` is invalid.
 
         """
-        res = requests.get(url=f'{DOMAIN}/.api/v1/mailing/{id}/', headers=self.HEADERS)
+        res = requests.get(url=f'{DOMAIN}/.api/v1/mailing/{id}/', headers=self._HEADERS)
         if not res.ok:
             if res.status_code in (403, 404):
                 raise exceptions.ForbiddenMailingError('The mailing for the specified id either does not exist or is not owned by the client\'s user')
@@ -250,7 +306,7 @@ class AMcardsClient:
             'offset': skip,
         } | (filters or {})
 
-        res = requests.get(url=f'{DOMAIN}/.api/v1/contact/', headers=self.HEADERS, params=params)
+        res = requests.get(url=f'{DOMAIN}/.api/v1/contact/', headers=self._HEADERS, params=params)
         if not res.ok:
             raise exceptions.AuthenticationError('Access token provided to client is unauthorized')
 
@@ -269,7 +325,7 @@ class AMcardsClient:
         :raises AuthenticationError: When the client's ``access_token`` is invalid.
 
         """
-        res = requests.get(url=f'{DOMAIN}/.api/v1/contact/{id}/', headers=self.HEADERS)
+        res = requests.get(url=f'{DOMAIN}/.api/v1/contact/{id}/', headers=self._HEADERS)
         if not res.ok:
             if res.status_code in (403, 404):
                 raise exceptions.ForbiddenContactError('The contact for the specified id either does not exist or is not owned by the client\'s user')
@@ -347,7 +403,7 @@ class AMcardsClient:
         if anniversary_month is not None: body |= {'anniversary_month': anniversary_month}
         if anniversary_day is not None: body |= {'anniversary_day': anniversary_day}
 
-        res = requests.post(url=f'{DOMAIN}/.api/v1/contact/', json=body, headers=self.HEADERS)
+        res = requests.post(url=f'{DOMAIN}/.api/v1/contact/', json=body, headers=self._HEADERS)
         if not res.ok:
             if res.status_code == 401:
                 raise exceptions.AuthenticationError('Access token provided to client is unauthorized')
@@ -362,7 +418,7 @@ class AMcardsClient:
         :raises AuthenticationError: When the client's ``access_token`` is invalid.
 
         """
-        res = requests.delete(url=f'{DOMAIN}/.api/v1/contact/{id}/', headers=self.HEADERS)
+        res = requests.delete(url=f'{DOMAIN}/.api/v1/contact/{id}/', headers=self._HEADERS)
         if not res.ok:
             if res.status_code in (403, 404):
                 raise exceptions.ForbiddenContactError('The contact for the specified id either does not exist or is not owned by the client\'s user')
@@ -586,7 +642,7 @@ class AMcardsClient:
         if message is not None:
             body |= {'message': message}
 
-        res = requests.post(f'{DOMAIN}/cards/open-card-form-oa/', json=body, headers=self.HEADERS)
+        res = requests.post(f'{DOMAIN}/cards/open-card-form-oa/', json=body, headers=self._HEADERS)
 
         # Check for errors
         match res.status_code:
@@ -725,7 +781,7 @@ class AMcardsClient:
             body['recipients'][0]['birth_day'] = shipping_address['birth_date'][-2:]
             body['recipients'][0]['birth_month'] = shipping_address['birth_date'][-5:-3]
 
-        res = requests.post(f'{DOMAIN}/campaigns/calculate-campaign-price/', json=body, headers=self.HEADERS)
+        res = requests.post(f'{DOMAIN}/campaigns/calculate-campaign-price/', json=body, headers=self._HEADERS)
 
         # Check for errors
         match res.status_code:
@@ -867,7 +923,7 @@ class AMcardsClient:
         if send_date is not None:
             body |= {'send_date': send_date}
 
-        res = requests.post(f'{DOMAIN}/campaigns/open-campaign-form/', json=body, headers=self.HEADERS)
+        res = requests.post(f'{DOMAIN}/campaigns/open-campaign-form/', json=body, headers=self._HEADERS)
 
         # Check for errors
         match res.status_code:
@@ -1032,7 +1088,7 @@ class AMcardsClient:
         if send_date is not None:
             body |= {'send_type': 'specific_date', 'send_date': send_date}
 
-        res = requests.post(f'{DOMAIN}/cards/open-mailing-form/', json=body, headers=self.HEADERS)
+        res = requests.post(f'{DOMAIN}/cards/open-mailing-form/', json=body, headers=self._HEADERS)
 
         # Check for errors
         match res.status_code:
